def standardize_name(endpoint):
    endpoint = endpoint.lower()
    endpoint = endpoint.replace("/", "_")
    endpoint = endpoint.replace("-", "_")
    return endpoint

def verify_endpoint(endpoint, df):
    import pandas as pd
    from json import loads
    
    endpoint_name = standardize_name(endpoint)

    default_json_file = f"{endpoint_name}.json"
    file_path = f"../defaults/types/{default_json_file}"
    import os
    this_dir, this_filename = os.path.split(__file__)
    DATA_PATH = os.path.join(this_dir, file_path)
    type_file = open(DATA_PATH)
    default_type = loads(type_file.read())
    type_file.close()
    
    default_df = pd.json_normalize(default_type, sep="_")
    df_columns_list = df.columns.values.tolist()
    default_df_columns_list = default_df.columns.values.tolist()
    # print(df_columns_list)
    # print(default_df_columns_list)

    error_count = 0
    for column in df_columns_list:
        try:
            default_df_columns_list.index(column)
        except Exception as e:
            print(f"Error trying to find {column} in default for endpoint {endpoint_name}. Please update JSON to include column. ")
            error_count += 1
    print(f"Endpoint {endpoint_name} has been verified. There are a total of {error_count} issues.")

    return default_df


def pull_full_dataframe(engine, st_creds_path, endpoint, titanpy_instance, start_date=None):

    import pandas as pd
    import logging
    from .database_tools import stage_then_merge

    print(f"Saving logs to {standardize_name(endpoint)}.log")
    logging.basicConfig(filename=f"atlas.log", force=True)
    print(f"________________ Getting data from {endpoint} _________________")

    tp = titanpy_instance

    query = {
        "tenant": tp.tenant_id,
        "from": start_date
    }

    result = tp.Get(endpoint,query=query)

    result_json = result.json()
    # print(result_json)
    data = result_json['data']
    has_more = result_json['hasMore']
    continue_from = result_json['continueFrom']
    total_length = 0

    df = pd.json_normalize(data, sep="_")
    verify_endpoint(endpoint, df)
    total_length += len(df)

    df["tenant_id"] = tp.tenant_id
    df_dict = {}
    df_dict[standardize_name(endpoint)] = df
    stage_then_merge(engine, df_dict)

    while has_more==True:
        query["from"] = continue_from
        logging.info(f"More data to request. Continue from code: {continue_from}")
        try:
            result = tp.Get(endpoint,query=query)
            while result.status_code != 200:
                if result.status_code == 401:
                    tp.Connect(cred_path = st_creds_path)
                    result = tp.Get(endpoint,query=query)
                    print(result.json())
                else:
                    from time import sleep
                    logging.info("Waiting 30 seconds due to rate limit")
                    sleep(30)
                    result = tp.Get(endpoint,query=query)
        except Exception as e:
            logging.error(e)
        result_json = result.json()
        data = result_json['data']
        has_more = result_json['hasMore']
        continue_from = result_json['continueFrom']
        df2 = pd.json_normalize(data, sep="_")
        df2["tenant_id"] = tp.tenant_id
        df_dict = {}
        df_dict[standardize_name(endpoint)] = df2
        stage_then_merge(engine, df_dict)
        total_length += len(df2)

        logging.info(f"Total Results ({endpoint}) so far: {total_length}")

def titanpy_dataframe(engine, endpoint_list, st_creds_path, start_date=None):
    
    from Titanpy.titanpy import Titanpy
    from threading import Thread

    tp = Titanpy()
    tp.Connect(cred_path = st_creds_path)

    if isinstance(endpoint_list, list):
        pass
    else:
        endpoint_list = [endpoint_list]

    for endpoint in endpoint_list:
        # pull_full_dataframe(engine,st_creds_path,endpoint,tp,start_date)
        Thread(target = pull_full_dataframe, args=(engine, st_creds_path,endpoint, tp, start_date)).start()



    


# This script contains functions that Atlas uses to created different types of source datasets and initializes the database using sqlalchemy.

def Default(sql_creds_path, st_creds_path, start_date = None):
    
    from .database_tools import create_engine, run_sql
    from .titanpy_tools import titanpy_dataframe
    
    endpoint_list = [
        # 'export/inventory-bills',
        'export/invoice-items',
        'export/invoices',
        'export/payments',
        'export/bookings',
        'export/customers',
        'export/customers/contacts',
        'export/leads',
        'export/locations',
        'export/locations/contacts',
        'export/appointment-assignments',
        'export/purchase-orders',
        'export/appointments',
        'export/job-canceled-logs',
        'export/jobs',
        'export/projects',
        'export/invoice-templates',
        'export/membership-types',
        'export/memberships',
        'export/recurring-service-events',
        'export/recurring-service-types',
        'export/recurring-services',
        'export/activity-codes',
        'export/gross-pay-items',
        'export/jobs/splits',
        'export/jobs/timesheets',
        'export/payroll-adjustments',
        'export/timesheet-codes',
        'export/business-units',
        'export/employees',
        'export/tag-types',
        'export/technicians',
    ]

    engine = create_engine(sql_creds_path)

    print("\n------------------ Initializing Default Atlas Schema -------------------")
    import os
    this_dir, this_filename = os.path.split(__file__)
    DATA_PATH = os.path.join(this_dir, "../sql/create_atlas_schema.sql")
    run_sql(engine, DATA_PATH)
    print("\nAtlas Schema Initialized")
    print("\n------------------ Getting Servicetitan Data ---------------------------\n")
    df_dict = titanpy_dataframe(engine, endpoint_list, st_creds_path, start_date=start_date)
    print("\n------- Succesfully Created/Updated Default Atlas Source Tables! -------\n")


if __name__ == "__main__":
    Default("./credentials/postgresql_credentials.json", "./credentials/servicetitan_credentials.json", start_date='2023-11-01')



def create_engine(creds_file_path):

    from sqlalchemy import create_engine
    from json import load

    try:
        creds_json = open(creds_file_path)
        creds_data = load(creds_json)
        creds_json.close()
    except Exception as e:
        print("There was an error with the requested credential path.")
        print(e)

    if creds_data['db_type'] == 'postgresql15':
        connection_string = f"postgresql://{creds_data['user']}:{creds_data['pass']}@{creds_data['host']}:{creds_data['port']}/{creds_data['dbname']}"
    
    engine = create_engine(connection_string)

    return engine


def run_sql(engine, file_path):
    # Will run a sql file starting from the files base directory.
    from sqlalchemy import text
    sql_file = open(file_path)
    sql = sql_file.read()
    sql_file.close()
    sqlcommands = sql.split(';')

    while("" in sqlcommands):
        sqlcommands.remove("")

    with engine.connect() as conn:
        if len(sqlcommands)>1:
            for command in sqlcommands:
                conn.execute(text(command))
            conn.commit()
        else:
            result = conn.execute(text(sql))
            conn.commit()
            return result
        
def clean_df_objects(df):
    for column in df:
        if df[column].dtype == "object":
            bad_characters = ['\x00','\t','\n','\r','\x0a']
            for char in bad_characters:
                try:
                    df[column] = df[column].str.replace(char,'', regex=True)
                except:
                    pass
    return df

def create_stage(engine,endpoint):
    import pandas as pd
    from os import path, getcwd
    from json import loads, dumps
    from .titanpy_tools import standardize_name
    from sqlalchemy import text

    endpoint_name = endpoint
    default_json_file = f"{endpoint_name}.json"
    file_path = f"../defaults/types/{default_json_file}"
    import os
    this_dir, this_filename = os.path.split(__file__)
    DATA_PATH = os.path.join(this_dir, file_path)
    type_file = open(DATA_PATH)
    default_type = loads(type_file.read())
    type_file.close()
    default_df = pd.json_normalize(default_type, sep='_', max_level=None)
    default_df["tenant_id"] = 0
    default_df = default_df[0:0]
    with engine.connect() as conn:
        
        default_df.to_sql(f"st_src_{endpoint_name}_stage", conn, if_exists='replace', schema='atlas', index=False)

def stage_then_merge(engine, df_dict):

    from sqlalchemy import text
    from pandas import read_sql
    import logging

    for name in df_dict:
        logging.basicConfig(filename=f"atlas.log", force=True)
        source_name = f"st_src_{name}"
        logging.info(f"Creating {source_name} staging table...")
        df = df_dict[name]
        df = clean_df_objects(df)
        df.dropna(how='all', axis=1, inplace=True)
        if len(df.index) == 0:
            logging.info(f"No data returned for endpoint {source_name}")
            pass
        else:
            with engine.connect() as conn:
                create_stage(engine, name)
                df.to_sql(f"{source_name}_stage", conn, if_exists='append', schema='atlas', index=False)

                conn.execute(text(f"CREATE TABLE IF NOT EXISTS atlas.{source_name} AS select * from atlas.{source_name}_stage"))
                conn.commit()

                df2 = read_sql(text(f"select * from atlas.{source_name}"), engine)
                df = read_sql(text(f"select * from atlas.{source_name}_stage"), engine)
                df_columns = ''.join(f'stage_table."{col}", ' for col in df.columns.values.tolist())
                df_columns = df_columns[:-2]
                df2_columns = ''.join(f'"{col}", ' for col in df2.columns.values.tolist())
                df2_columns = df2_columns[:-2]
                update_string = ''.join(f'"{col}" = stage_table."{col}", ' for col in df2.columns.values.tolist())
                update_string = update_string[:-2]

                logging.info(f"Merging into source table...")
                conn.execute(text(f"merge into atlas.{source_name} source_table using atlas.{source_name}_stage stage_table on source_table.id = stage_table.id when matched then update set {update_string} when not matched then insert ({df2_columns}) values ({df_columns});"))
                conn.commit()
                logging.info(f"Completed merge!")



def load_credentials(cred_path):

    import requests
    from json import load

    # opens and stores json in creds_data variable
    try:
        creds_json = open(cred_path)
        creds_data = load(creds_json)
        creds_json.close()
    except Exception as e:
        print("There was an error with the requested credential path.")
        print(e)

    # set up variables
    auth_url = 'https://auth.servicetitan.io/connect/token'
    form = {
        "grant_type": "client_credentials",
        "client_id": creds_data["CLIENT_ID"],
        "client_secret": creds_data["CLIENT_SECRET"]
    }
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
    }
    # make request to return access_token
    try:
        request = requests.post(auth_url,data=form, headers=headers)
        if request.status_code == 200:
            creds_data["ACCESS_TOKEN"] = request.json()["access_token"]
            print(f"Access token successfully retrieved for Tenant {creds_data['TENANT_ID']}")
            return creds_data
        else:
            print(f"There was an error when requesting access token. Status returned: {request.status_code}")
    except Exception as e:
        print("There was an unknown error when requesting access token.")
        print(e)
